#ifndef __orb_protocol__
#define __orb_protocol__

#include "stock.h"
#include "bns.h"
#define ORBMAGIC             0x6d62726f // 'orbm' 
#define ORBSRCNAME_SIZE 64
#define ORBWHO_SIZE 24
#define ORBWHAT_SIZE 128
#define ORBHOST_SIZE 88
#define ORBVERSION_SIZE 128
#define ORBUNAME_SIZE 88
#define SELECTMAX 96
#define ORBSYNC "orbm"
#define NORBSYNC 4  // = strlen(ORBSYNC) 
#define ORBPUT               1
#define ORBGET               2
#define ORBREAP              3
#define ORBSTAT              4
#define ORBCLIENTS	     5
#define ORBSOURCES	     6
#define ORBSEEK              7
#define ORBTELL              8
#define ORBSELECT            9
#define ORBREJECT	     10
#define ORBCLOSE             11
#define ORBAFTER             12
#define ORBBEFORE            13
#define ORBPING              14
#define ORBHDR               15
#define ORBSETPRI	     16 
#define ORBHALT              17
#define ORBSETLOGGING	     18
#define ORBERROR  	     19
#define ORBOPEN 	     20
#define ORBKILL 	     21
#define ORBID		     23
#define ORBRUP		     24
#define ORBPUTL              25
#define ORBGETL              26
#define ORBREAPL             27
#define ORBSTASHSELECT       28
#define ORBGETSTASH          29
#define ORBPUTX              30
#define ORBPUTXL             31
#define ORB_ENDREAP         -32
#define ORBOPENX             33
#define ORBOPENY             34
#define ORBINVITE	     35
#define ORBIP		     36
#define ORBDLID		     37
#define ORBLAGVALUE	     38
#define ORBRANGE	     39
#define STASH_ALL               0
#define STASH_ONLY              1
#define NO_STASH                2
#define MAX_RETURNED_SELECTION   96
#define MAX_SELECTION_STRING  16384
#define CHALLENGE_SIZE       12
#define MESSAGE_SIZE         (2+ORBWHAT_SIZE+ORBWHO_SIZE+CHALLENGE_SIZE)
#define CHALLENGE_MSG_SIZE   (CHALLENGE_SIZE+2)
#define ORBSRCNAME_MSG_SIZE  (ORBSRCNAME_SIZE+2)
#define SG_MAX_SIGNATURE_MSG_SIZE  (SG_MAX_SIGNATURE+2)
#include "Pktpkg.h"

typedef struct Orbclient {
    double     started ;  /* time client was connected   */
    double     lastpkt ;  /* time of latest packet to/from client   */
    int        pktid ;  /* current packet id   */
    unsigned int packets ;  /* number of packets to/from this client   */
    unsigned int bytes ;  /* packet bytes to/from this client   */
    unsigned int read ;  /* #bytes read   */
    unsigned int written ;  /* #bytes written   */
    unsigned int nreads ;  /* #read calls   */
    unsigned int nwrites ;  /* #write calls   */
    unsigned int nrequests ;  /* # requests from client   */
    float      lag ;  /* current lag value  */
    int        lastrequest ;  /* # last request from client   */
    float      lagprev ;  /* previous lag value for reap threads  */
    int        errors ;  /* # errors from this client   */
    int        nselect ;  /* number of characters in select   */
    char       select[MAX_RETURNED_SELECTION] ;  /* current selection expression   */
    int        nreject ;  /* number of characters in reject   */
    char       reject[MAX_RETURNED_SELECTION] ;  /* current rejection expression   */
    int        fd ;  /* port to which client is connected   */
    unsigned int thread ;  /* thread to which client is connected   */
    char       address[4] ;  /* ip address of client machine   */
    int        port ;  /* client port on client machine   */
    char       who[ORBWHO_SIZE] ;  /* user running client   */
    char       host[ORBHOST_SIZE] ;  /* hostname of client machine   */
    double     prev_time ;  /* previous time  */
    unsigned int prev_packets ;  /* previous # of packets  */
    unsigned int prev_bytes ;  /* previous # of bytes  */
    unsigned int nstash ;  /* # of stash packets  */
    unsigned int prev_stash ;  /* previous # of stash packets  */
    float      utilization ;  /* estimate of cpu usage  */
    float      pkts_per_sec ;  /* packets/second  */
    float      bytes_per_sec ;  /* bytes/second  */
    float      stash_per_sec ;  /* stash packets / second  */
    char       what[ORBWHAT_SIZE] ;  /* command line for client machine   */
    int        pid ;  /* pid of client   */
    char       perm ;  /* permission (r or w)   */
} Orbclient ;

typedef struct OrbOpenString {
    int        handshake ;  /* another sanity check   */
    char       who[ORBWHO_SIZE] ;  /* user running client   */
    char       what[ORBWHAT_SIZE] ;  /* command line for client machine   */
    int        pid ;  /* pid of client   */
    char       perm ;  /* permission (r or w)   */
} OrbOpenString ;

typedef struct Orbstat {
    unsigned int maxdata ;  /* max data bytes  */
    int        maxpktid ;  /* max #packets  */
    int        maxsrc ;  /* maximum number of sources  */
    int        nclients ;  /* number of clients  */
    int        nsources ;  /* number of sources  */
    double     started ;  /* start time of orb server  */
    double     when ;  /* time when this status was taken  */
    int        pid ;  /* pid for orbserver  */
    char       address[4] ;  /* ip address   */
    int        port ;  /* port number for server requests  */
    int        opens ;  /* number of opens  */
    int        closes ;  /* number of closes  */
    int        rejected ;  /* number of failed opens   */
    int        errors ;  /* number of threads exiting with errors  */
    char       host[ORBUNAME_SIZE] ;  /* hostname:cwd where orbserver runs  */
    float      utilization ;  /* estimate of overall cpu utilization  */
    float      ipktrate ;  /* input packet rate  */
    float      opktrate ;  /* output packet rate  */
    float      ibyterate ;  /* input byte rate  */
    float      obyterate ;  /* output byte rate  */
    float      istashrate ;  /* input stash rate  */
    float      ostashrate ;  /* output stash rate  */
    float      maxlag ;  /* maximum current lag  */
    float      howlong ;  /* snooze amount * 16  */
    short      flags ;  /* bit 0 = throttling enabled  */
    unsigned char reaping ;  /* how many clients are reaping  */
    unsigned char stalled ;  /* how many reap clients are stalled  */
    char       who[ORBWHO_SIZE] ;  /* who ran the orbserver  */
    char       version[ORBVERSION_SIZE] ;  /* version code  */
    double     orb_start ;  /* last time buffer was initialized  */
    int        messages ;  /* global flag to print all messages if != 0   */
    int        connections ;  /* global flag to print all connection open and closes if != 0  */
} Orbstat ;

typedef struct Orbrequest {
    int        what ;  /* request code   */
    int        athread ;
    OrbOpenString *hello ;
    int        sZhello ;
    char       challenge[CHALLENGE_SIZE] ;
    char       signer[ORBSRCNAME_SIZE] ;
    char       signature[SG_MAX_SIGNATURE] ;
    int        nselect ;  /* characters in select + 1  */
    char       *select ;
    int        sZselect ;
    int        nreject ;  /* characters in reject + 1  */
    char       *reject ;
    int        sZreject ;
    char       srcname[ORBSRCNAME_SIZE] ;
    int        pktcode ;
    int        logflag ;
    double     time ;
    Pktpkg     *pkg ;
    int        sZpkg ;
    int        newpri ;
    int        thread ;
    int        stashflag ;
    char       ipstring[INET6_ADDRSTRLEN] ;
    double     reap_maxtime ;
    int        reap_maxbytes ;
    int        reap_maxpkts ;
    char       collector[ORBSRCNAME_SIZE] ;
} Orbrequest ;
#define ERRMAX 256

typedef struct Orbresponse {
    int        what ;  /* echo request   */
    int        result ;  /* result code   */
    int        active ;  /* 1 if thread is active, 0 otherwise  */
    double     orb_start ;  /* when orb ring buffer was last reset   */
    double     server_start ;  /* when orbserver was started   */
    int        latest_pktid ;  /* latest packet id  */
    char       challenge[CHALLENGE_SIZE] ;
    int        thread ;
    int        version ;
    int        nselections ;
    Pktpkg     *pkg ;
    int        sZpkg ;
    int        pktid ;
    Orbstat    orbstat ;
    int        stashflag ;
    double     client_when ;
    int        nclient ;
    Orbclient  *client ;
    int        sZclient ;
    double     source_when ;
    int        nsource ;
    Orbsrc     *source ;
    int        sZsource ;
    int        ndlid ;
    char       *dlid ;
    int        sZdlid ;
    double     lagvalue ;
    int        maxpktid ;
    int        eldest ;
    int        newest ;
    int        range ;
    int        nerr ;
    char       errmsg[ERRMAX] ;
} Orbresponse ;
#define freeBns bnsfree
#define initBns bnsinit
#define freeFILE fclose

typedef struct Orbbones {
    double     lastpkttime ;  /* time of last packet from reap or get  */
    double     maxpkttime ;  /* most recent time from any reap  */
    double     orb_start ;  /* ring buffer initialization time from latest open  */
    int        lastpktid ;  /* last pktid from reap, get, orbseek, or orbafter  */
} Orbbones ;

typedef struct Orbfile {
    double     opentime ;  /* last time at which socket was opened  */
    double     server_start ;  /* orbserver start time from latest open  */
    char       *name ;  /* orbserver name  */
    int        sZname ;
    char       signer[ORBSRCNAME_SIZE] ;  /* signer for non-ip based security  */
    Bns        *in ;  /* buffered input socket stream  */
    int        sZin ;
    Bns        *out ;  /* buffered output socket stream  */
    int        sZout ;
    Orbrequest *req ;  /* request sent to server  */
    int        sZreq ;
    Orbresponse *reply ;  /* reply from server  */
    int        sZreply ;
    Orbbones   bones ;  /* current read position information  */
    int        resend ;  /* number of packets to resend when orbput fails  */
    int        count ;  /* count of orbput_nolag packets since check  */
    int        range ;  /* range of packets on orbserver for orbput_nolag  */
    int        perm ;  /* initial open flag = 'r' or 'w'  */
    int        reaped ;  /* if !=0, currently reaping packets  */
    int        reopen ;  /* if !=0, reopen socket if an error occurs  */
    int        fd ;  /* file descriptor  */
    int        once ;  /* flag for reopen complaints  */
    int        first_nd_failure ;  /* separate timeout counter for orbreap_nd  */
    int        error ;  /* if !=0, socket has returned an error  */
    int        nonblocking ;  /* if !=0, has been set to non-blocking i/o  */
    int        is_forbserver ;  /* flag for forbserver program: no timeouts, no reopens */
    int        cnt ;  /* count of read timeouts with no data  */
    int        prevlast ;  /* if != -1, has previous last_pktid for remote orb  */
} Orbfile ;

#ifdef  __cplusplus
extern "C" {
#endif

extern Orbsrc *newOrbsrc() ;
extern Orbclient *newOrbclient() ;
extern OrbOpenString *newOrbOpenString() ;
extern Orbstat *newOrbstat() ;
extern Orbrequest *newOrbrequest() ;
extern Orbresponse *newOrbresponse() ;
extern Orbbones *newOrbbones() ;
extern Orbfile *newOrbfile() ;
extern void freeOrbsrc (Orbsrc *self) ;
extern void freeOrbclient (Orbclient *self) ;
extern void freeOrbOpenString (OrbOpenString *self) ;
extern void freeOrbstat (Orbstat *self) ;
extern void freeOrbrequest (Orbrequest *self) ;
extern void freeOrbresponse (Orbresponse *self) ;
extern void freeOrbbones (Orbbones *self) ;
extern void freeOrbfile (Orbfile *self) ;
extern void initOrbsrc (Orbsrc *self) ;
extern void initOrbclient (Orbclient *self) ;
extern void initOrbOpenString (OrbOpenString *self) ;
extern void initOrbstat (Orbstat *self) ;
extern void initOrbrequest (Orbrequest *self) ;
extern void initOrbresponse (Orbresponse *self) ;
extern void initOrbbones (Orbbones *self) ;
extern int gbnsOrbsrc(Bns *bns, Orbsrc *self) ;
extern int gbnsOrbclient(Bns *bns, Orbclient *self) ;
extern int gbnsOrbOpenString(Bns *bns, OrbOpenString *self) ;
extern int gbnsOrbstat(Bns *bns, Orbstat *self) ;
extern int gbnsOrbrequest(Bns *bns, Orbrequest *self) ;
extern int gbnsOrbresponse(Bns *bns, Orbresponse *self) ;
extern int gbnsOrbbones(Bns *bns, Orbbones *self) ;
extern int pbnsOrbsrc(Bns *bns, Orbsrc *self) ;
extern int pbnsOrbclient(Bns *bns, Orbclient *self) ;
extern int pbnsOrbOpenString(Bns *bns, OrbOpenString *self) ;
extern int pbnsOrbstat(Bns *bns, Orbstat *self) ;
extern int pbnsOrbrequest(Bns *bns, Orbrequest *self) ;
extern int pbnsOrbresponse(Bns *bns, Orbresponse *self) ;
extern int pbnsOrbbones(Bns *bns, Orbbones *self) ;
extern void pdbgOrbsrc(FILE *file, char *msg, Orbsrc *self, int verbose, int indent) ;
extern void pdbgOrbclient(FILE *file, char *msg, Orbclient *self, int verbose, int indent) ;
extern void pdbgOrbOpenString(FILE *file, char *msg, OrbOpenString *self, int verbose, int indent) ;
extern void pdbgOrbstat(FILE *file, char *msg, Orbstat *self, int verbose, int indent) ;
extern void pdbgOrbrequest(FILE *file, char *msg, Orbrequest *self, int verbose, int indent) ;
extern void pdbgOrbresponse(FILE *file, char *msg, Orbresponse *self, int verbose, int indent) ;
extern void pdbgOrbbones(FILE *file, char *msg, Orbbones *self, int verbose, int indent) ;
extern void pdbgOrbfile(FILE *file, char *msg, Orbfile *self, int verbose, int indent) ;

#ifdef  __cplusplus
}
#endif

#endif
